@@ -1,3 +1,14 @@
+"""
+image utils.
+
+Library format cheat sheet:
+
+Library     Dim Order       Channel Order       Value Range     Type
+Pillow                      R, G, B, A          0-255           PIL.Image.Image
+OpenCV                      B, G, R, A          0-255           np.ndarray
+Torch       (B), C, H, W    R, G, B             -1.0-1.0        torch.Tensor
+
+"""
 from typing import Sequence
 
 import numpy as np
@@ -34,14 +45,31 @@ def pillow_fit_image_within(
     return image
 
 
-def pillow_img_to_torch_image(img: PIL.Image.Image):
-    img = img.convert("RGB")
+def pillow_img_to_torch_image(img: PIL.Image.Image, convert="RGB"):
+    if convert:
+        img = img.convert(convert)
     img = np.array(img).astype(np.float32) / 255.0
+    # b, h, w, c => b, c, h, w
     img = img[None].transpose(0, 3, 1, 2)
     img = torch.from_numpy(img)
     return 2.0 * img - 1.0
 
 
+def pillow_mask_to_latent_mask(mask_img: PIL.Image.Image, downsampling_factor):
+    mask_img = mask_img.resize(
+        (
+            mask_img.width // downsampling_factor,
+            mask_img.height // downsampling_factor,
+        ),
+        resample=Image.Resampling.LANCZOS,
+    )
+
+    mask = np.array(mask_img).astype(np.float32) / 255.0
+    mask = mask[None, None]
+    mask = torch.from_numpy(mask)
+    return mask
+
+
 def pillow_img_to_opencv_img(img: PIL.Image.Image):
     open_cv_image = np.array(img)
     # Convert RGB to BGR
@@ -49,22 +77,58 @@ def pillow_img_to_opencv_img(img: PIL.Image.Image):
     return open_cv_image
 
 
-def model_latents_to_pillow_imgs(latents: torch.Tensor) -> Sequence[PIL.Image.Image]:
-    from classy_imaginary.model_manager import get_current_diffusion_model  # noqa
+def torch_image_to_openvcv_img(img: torch.Tensor):
+    img = (img + 1) / 2
+    img = img.detach().cpu().numpy()
+    # assert there is only one image
+    assert img.shape[0] == 1
+    img = img[0]
+    img = img.transpose(1, 2, 0)
+    img = (img * 255).astype(np.uint8)
+    # RGB to BGR
+    img = img[:, :, ::-1]
+    return img
 
+
+def torch_img_to_pillow_img(img_t: torch.Tensor):
+    img_t = img_t.to(torch.float32).detach().cpu()
+    if len(img_t.shape) == 3:
+        img_t = img_t.unsqueeze(0)
+    if img_t.shape[0] != 1:
+        raise ValueError("Only batch size 1 supported")
+    if img_t.shape[1] == 1:
+        colorspace = "L"
+    elif img_t.shape[1] == 3:
+        colorspace = "RGB"
+    else:
+        raise ValueError(
+            f"Unsupported colorspace. {img_t.shape[1]} channels in {img_t.shape} shape"
+        )
+    img_t = rearrange(img_t, "b c h w -> b h w c")
+    img_t = torch.clamp((img_t + 1.0) / 2.0, min=0.0, max=1.0)
+    img_np = (255.0 * img_t).cpu().numpy().astype(np.uint8)[0]
+    if colorspace == "L":
+        img_np = img_np[:, :, 0]
+    return Image.fromarray(img_np, colorspace)
+
+
+def model_latent_to_pillow_img(latent: torch.Tensor) -> PIL.Image.Image:
+    from imaginairy.model_manager import get_current_diffusion_model  # noqa
+
+    if len(latent.shape) == 3:
+        latent = latent.unsqueeze(0)
+    if latent.shape[0] != 1:
+        raise ValueError("Only batch size 1 supported")
     model = get_current_diffusion_model()
-    latents = model.decode_first_stage(latents)
-    latents = torch.clamp((latents + 1.0) / 2.0, min=0.0, max=1.0)
-    imgs = []
-    for latent in latents:
-        latent = 255.0 * rearrange(latent.cpu().numpy(), "c h w -> h w c")
-        img = Image.fromarray(latent.astype(np.uint8))
-        imgs.append(img)
-    return imgs
+    img_t = model.decode_first_stage(latent)
+    return torch_img_to_pillow_img(img_t)
+
+
+def model_latents_to_pillow_imgs(latents: torch.Tensor) -> Sequence[PIL.Image.Image]:
+    return [model_latent_to_pillow_img(latent) for latent in latents]
 
 
 def pillow_img_to_model_latent(model, img, batch_size=1, half=True):
-    # init_image = pil_img_to_torch(img, half=half).to(device)
     init_image = pillow_img_to_torch_image(img).to(get_device())
     init_image = repeat(init_image, "1 ... -> b ...", b=batch_size)
     if half:
